import os
from os import environ

import otree.settings
from django.conf.global_settings import STATICFILES_STORAGE  # noqa

import dj_database_url

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

PRJ_DIR = os.path.dirname(BASE_DIR)

DEBUG = True

ADMIN_PASSWORD = 'otree'
SECRET_KEY = 'zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz'

AUTH_LEVEL = ''


environ['DATABASE_URL'] = 'postgres://postgres@localhost/django_db'


DATABASES = {
    'default': dj_database_url.config(
        default='sqlite:///' + os.path.join(BASE_DIR, 'db.sqlite3')
    )
}
'''

# 2017-06-13: why did i have it hardcoded previously? maybe because SQLite is faster?
# 2017-07-16: because postgres gives me an error, even after i reset the DB:
#    django.db.utils.ProgrammingError: relation "otree_session" does not exist

# 2017-06-13: why did i have it hardcoded previously? maybe because SQLite is faster?


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        #'TEST': {
        #    'NAME': 'testdb.sqlite3'
        #}
    }
}
'''

CREATE_DEFAULT_SUPERUSER = True
ADMIN_USERNAME = 'admin'

# should keep these so i can run my MTurk tests
AWS_ACCESS_KEY_ID = environ.get('AWS_ACCESS_KEY_ID')
AWS_SECRET_ACCESS_KEY = environ.get('AWS_SECRET_ACCESS_KEY')

# e.g. EUR, CAD, GBP, CHF, CNY, JPY
REAL_WORLD_CURRENCY_CODE = 'EUR'
USE_POINTS = False


# e.g. en-gb, de-de, it-it, fr-fr.
# see: https://docs.djangoproject.com/en/1.7/topics/i18n/
LANGUAGE_CODE = 'en'

INSTALLED_APPS = [
    'otree',
    'raven.contrib.django.raven_compat',
    'tests',
]

SESSION_CONFIG_DEFAULTS = {
    'real_world_currency_per_point': 0.01,
    'participation_fee': 10.00,
    'doc': "",
    'use_browser_bots': False,
}


SESSION_CONFIGS = [
    {
        'name': 'simple',
        'display_name': "Simple Game",
        'num_demo_participants': 1,
        'app_sequence': ['tests.simple'],
    },
    {
        'name': 'saving',
        'display_name': "Saving",
        'num_demo_participants': 1,
        'participation_fee': 9.99,
        'real_world_currency_per_point': 0.02,
        'app_sequence': ['tests.saving'],
        'treatment': 'blue',

    },
    {
        'name': 'misc_3p',
        'display_name': "Multi Player Game",
        'num_demo_participants': 3,
        'app_sequence': ['tests.misc_3p'],
    },
    {
        "name": 'two_simple_games',
        "display_name": "2 Simple Games",
        "num_demo_participants": 1,
        "app_sequence": ['tests.simple', 'tests.saving'],
    },
    {
        'name': 'skip_many',
        'display_name': "skip many",
        'num_demo_participants': 2,
        'app_sequence': ['tests.skip_many'],
    },
    {
        'name': 'rounds',
        'num_demo_participants': 4,
        'app_sequence': ['tests.rounds'],
    },
    {
        'name': 'advance_slowest',
        'num_demo_participants': 2,
        'app_sequence': ['tests.advance_slowest'],
    },
    {
        'name': 'advance_slowest_wait',
        'num_demo_participants': 2,
        'app_sequence': ['tests.advance_slowest_wait'],
    },
    {
        'name': 'wait_page',
        'num_demo_participants': 2,
        'app_sequence': ['tests.wait_page'],
    },
    {
        'name': 'skip_wait_page',
        'num_demo_participants': 2,
        'app_sequence': ['tests.skip_wait_page'],
    },
    {
        'name': 'waitpage_set_field',
        'num_demo_participants': 4,
        'app_sequence': ['tests.waitpage_set_field'],
    },
    {
        'name': 'waitpage_misuse',
        'num_demo_participants': 2,
        'app_sequence': ['tests.waitpage_misuse'],
    },
    {
        'name': 'skip_waitpage_lookahead',
        'num_demo_participants': 2,
        'app_sequence': ['tests.skip_waitpage_lookahead'],
        'use_browser_bots': True,
    },
    {
        'name': 'form_validation',
        'num_demo_participants': 1,
        'app_sequence': ['tests.form_validation'],
    },
    {
        'name': 'export',
        'num_demo_participants': 1,
        'app_sequence': ['tests.export'],
    },
    {
        'name': 'bots_raise',
        'num_demo_participants': 2,
        'app_sequence': ['tests.bots_empty', 'tests.bots_raise'],
    },
    {
        'name': 'group_by_arrival_time',
        'num_demo_participants': 6,
        'app_sequence': ['tests.group_by_arrival_time'],
        'use_browser_bots': True,
    },
    {
        'name': 'group_by_arrival_time_round1',
        'num_demo_participants': 6,
        'app_sequence': ['tests.group_by_arrival_time_round1'],
        'use_browser_bots': True,
    },
    {
        'name': 'group_by_arrival_time_custom',
        'num_demo_participants': 6,
        'app_sequence': ['tests.group_by_arrival_time_custom'],
        'use_browser_bots': True,
    },
    {
        'name': 'group_by_arrival_time_heartbeat',
        'num_demo_participants': 4,
        'app_sequence': ['tests.group_by_arrival_time_heartbeat'],
    },
    {
        'name': 'bots_check_html',
        'num_demo_participants': 1,
        'app_sequence': ['tests.bots_check_html'],
    },
    {
        'name': 'bots_bad_post',
        'num_demo_participants': 1,
        'app_sequence': ['tests.bots_bad_post'],
    },
    {
        'name': 'bots_cases',
        'num_demo_participants': 1,
        'app_sequence': ['tests.bots_cases'],
    },
    {
        'name': 'templates_app',
        'num_demo_participants': 1,
        'app_sequence': ['tests.templates_app'],
    },
    {
        'name': 'i18n',
        'num_demo_participants': 1,
        'app_sequence': ['tests.i18n'],
    },
    {
        'name': 'admin_report',
        'num_demo_participants': 1,
        'app_sequence': ['tests.admin_report'],
    },
    {
        'name': 'two_rounds_1p',
        'num_demo_participants': 1,
        'app_sequence': ['tests.two_rounds_1p'],
        'real_world_currency_per_point': 0.5,
        'participation_fee': 1.25,
    },
    {
        'name': 'timeout_submission',
        'num_demo_participants': 1,
        'app_sequence': ['tests.timeout_submission'],
    },
    {
        'name': 'constants',
        'num_demo_participants': 1,
        'app_sequence': ['tests.constants'],
    },
    {
        'name': 'mturk',
        'num_demo_participants': 1,
        'app_sequence': ['tests.simple'],
        'participation_fee': 0.4,
        'mturk_hit_settings': {
            'keywords': ['easy', 'bonus', 'choice', 'study'],
            'title': 'Title for your experiment',
            'description': 'Description for your experiment',
            'frame_height': 500,
            'preview_template': 'global/MTurkPreview.html',
            'minutes_allotted_per_assignment': 60,
            'expiration_hours': 7 * 24,  # 7 days
            # to prevent retakes
            'grant_qualification_id': 'ATESTQUAL',
            'qualification_requirements': []
        }
    },
]


ROOM_DEFAULTS = {}

ROOM_WITH_LABELS_NAME = 'with_labels'
ROOM_WITHOUT_LABELS_NAME = 'without_labels'

ROOMS = [
    {
        'name': ROOM_WITH_LABELS_NAME,
        'display_name': 'with_labels',
        'participant_label_file': 'tests/participant_labels.txt',
        'use_secure_urls': False
    },
    {
        'name': ROOM_WITHOUT_LABELS_NAME,
        'display_name': 'Anonymous',
    },
]

BOTS_CHECK_HTML = False

otree.settings.augment_settings(globals())


CHANNEL_LAYERS = globals()['CHANNEL_LAYERS']
REDIS_URL = globals()['REDIS_URL']

# i could put this in otree-core, but i don't want to mess anything up,
# e.g. bots
CHANNEL_LAYERS['default']["TEST_CONFIG"] = {"hosts": [REDIS_URL]}




# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants
import random
from six.moves import range


class FieldOnOtherPlayer(Page):

    def is_displayed(self):
        return self.player.id_in_group == 1

    def before_next_page(self):
        for p in self.group.get_players():
            p.from_other_player = 1
        in_all_rounds = self.player.in_all_rounds()
        assert ([p.subsession.round_number for p in in_all_rounds] ==
                list(range(1, self.subsession.round_number + 1)))
        assert in_all_rounds[-1].from_other_player == 1


class PickWinner(WaitPage):

    def after_all_players_arrive(self):
        # testing that this code only gets executed once
        winner = random.choice(self.group.get_players())
        winner.is_winner = True

        for p in self.group.get_players():
            p.participant.vars['set in wait page'] = True


class ResultsWaitPage(WaitPage):

    def after_all_players_arrive(self):
        self.group.set_payoffs()

        if self.subsession.round_number == Constants.num_rounds:
            this_group_players = self.group.get_players()
            participants = set(
                p.participant for p in this_group_players
            )

            for i, group in enumerate(self.group.in_previous_rounds()):
                assert group.subsession.round_number == i + 1
                players = group.get_players()
                assert len(players) == len(this_group_players)

                assert all(
                    p.participant in participants
                    for p in players
                )


class AllGroupsWaitPage(WaitPage):
    wait_for_all_groups = True

    def after_all_players_arrive(self):

        if self.subsession.round_number == Constants.num_rounds:
            for p in self.subsession.get_players():
                if p.id_in_group == 1:
                    p.group_id_before_p1_switch = p.group.id
            self.subsession.shuffle_p1()

        for p in self.subsession.get_players():
            p.in_all_groups_wait_page = 5.0
        for g in self.subsession.get_groups():
            g.in_all_groups_wait_page = 5.0


class Results(Page):

    def vars_for_template(self):
        assert self.player.from_other_player == 1
        assert self.player.in_all_groups_wait_page == 5.0
        assert self.group.in_all_groups_wait_page == 5.0

        assert self.player.participant.vars['set in wait page']

        if self.subsession.round_number == Constants.num_rounds:
            for p in self.subsession.get_players():
                if p.id_in_group == 1:

                    assert p.group_id_before_p1_switch != p.group.id

page_sequence = [
    FieldOnOtherPlayer,
    PickWinner,
    ResultsWaitPage,
    AllGroupsWaitPage,
    Results
]



from django.core.urlresolvers import reverse
from otree.session import create_session
from .utils import TestCase
import splinter
from django.conf import settings
from otree.models.participant import Participant

URL_ADMIN_LABELS = reverse('RoomWithoutSession', args=[settings.ROOM_WITH_LABELS_NAME])
URL_ADMIN_NO_LABELS = reverse('RoomWithoutSession', args=[settings.ROOM_WITHOUT_LABELS_NAME])
URL_PARTICIPANT_LABELS = reverse('AssignVisitorToRoom', args=[settings.ROOM_WITH_LABELS_NAME])
URL_PARTICIPANT_NO_LABELS = reverse('AssignVisitorToRoom', args=[settings.ROOM_WITHOUT_LABELS_NAME])

LABEL_REAL = 'JohnSmith'
LABEL_FAKE = 'NotInParticipantLabelsFile'


def add_label(base_url, label):
    return base_url + '?participant_label={}'.format(label)


class RoomTestCase(TestCase):

    def setUp(self):
        self.browser = splinter.Browser('django') # type: splinter.Browser

    def get(self, url):
        self.browser.visit(url)


class TestRoomWithoutSession(RoomTestCase):

    def test_open_admin_links(self):
        urls = [
            reverse('Rooms'),
            URL_ADMIN_LABELS,
            URL_ADMIN_NO_LABELS,
        ]

        for url in urls:
            self.get(url)

    def test_open_participant_links(self):

        WAITING_STR = 'Waiting for your session to begin'

        br = self.browser

        self.get(URL_PARTICIPANT_LABELS)
        self.assertIn('Please enter your participant label', br.html)

        self.get(
            add_label(URL_PARTICIPANT_LABELS, LABEL_REAL)
        )

        self.assertIn(WAITING_STR, br.html)

        self.get(add_label(URL_PARTICIPANT_LABELS, LABEL_FAKE))
        self.assertEqual(br.status_code, 404)


        self.get(URL_PARTICIPANT_NO_LABELS)
        self.assertIn(WAITING_STR, br.html)

        self.get(add_label(URL_PARTICIPANT_NO_LABELS, LABEL_REAL))
        self.assertIn(WAITING_STR, br.html)

    def test_ping(self):
        pass


class TestRoomWithSession(RoomTestCase):
    def _visited_count(self, session):
        return session.participant_set.filter(visited=True).count()

    def visited_count_labels(self):
        return self._visited_count(self.session_with_labels)

    def visited_count_no_labels(self):
        return self._visited_count(self.session_without_labels)

    def setUp(self):
        super().setUp()
        self.session_with_labels = create_session(
            'simple',
            # make it 6 so that we can test if the participant is reassigned
            # if they open their start link again (1/6 chance)
            num_participants=6,
            room_name=settings.ROOM_WITH_LABELS_NAME,
        )

        self.session_without_labels = create_session(
            'simple',
            num_participants=2,
            room_name=settings.ROOM_WITHOUT_LABELS_NAME,
        )

    def tearDown(self):
        for url in [URL_ADMIN_LABELS, URL_ADMIN_NO_LABELS]:
            self.get(url)
            button = self.browser.find_by_id('close-room')
            button.click()

        self.get(URL_ADMIN_LABELS)
        self.assertIn('room_without_session', self.browser.url)

    def test_open_admin_links(self):
        self.get(URL_ADMIN_LABELS)
        self.assertIn('room_with_session', self.browser.url)

    def test_without_label(self):

        self.get(URL_PARTICIPANT_NO_LABELS)
        self.get(URL_PARTICIPANT_NO_LABELS)

        # should use a cookie, so it remembers the participant
        self.assertEqual(self.visited_count_no_labels(), 1)

        self.get(add_label(URL_PARTICIPANT_NO_LABELS, LABEL_REAL))
        self.assertEqual(self.visited_count_no_labels(), 2)

    def test_with_label(self):
        br = self.browser

        self.get(URL_PARTICIPANT_LABELS)
        self.assertIn('Please enter your participant label', br.html)

        self.get(add_label(URL_PARTICIPANT_LABELS, LABEL_FAKE))
        self.assertEqual(br.status_code, 404)
        self.assertEqual(self.visited_count_labels(), 0)

        # make sure reopening the same link assigns you to same participant
        for i in range(2):
            self.get(add_label(URL_PARTICIPANT_LABELS, LABEL_REAL))

        visited_participants = Participant.objects.filter(visited=True)
        self.assertEqual(len(visited_participants), 1)

        participant = visited_participants[0]

        self.assertEqual(participant.label, LABEL_REAL)

    def test_delete_session_in_room(self):
        pass

    def test_session_start_links_room(self):
        pass



# This file is auto-generated.
# It's used to aid autocompletion in code editors.

import otree.api
from .. import models


class Page(otree.api.Page):
    def z_autocomplete(self):
        self.subsession = models.Subsession()
        self.group = models.Group()
        self.player = models.Player()


class WaitPage(otree.api.WaitPage):
    def z_autocomplete(self):
        self.subsession = models.Subsession()
        self.group = models.Group()


class Bot(otree.api.Bot):
    def z_autocomplete(self):
        self.subsession = models.Subsession()
        self.group = models.Group()
        self.player = models.Player()



#!/usr/bin/env python

# run the worker to enforce page timeouts
# even if the user closes their browser
from huey.contrib.djhuey.management.commands.run_huey import (
    Command as HueyCommand
)


class Command(HueyCommand):
    def handle(self, *args, **options):
        # clear any tasks in Huey DB, so they don't pile up over time,
        # especially if you run the server without the timeoutworker to consume
        # the tasks.
        # this code is also in asgi.py. it should be in both places,
        # to ensure the database is flushed in all circumstances.
        from huey.contrib.djhuey import HUEY
        HUEY.flush()
        super(Command, self).handle(*args, **options)



from django.core.management import call_command
from django.template import Context
from django.template import Template
from django.template import TemplateSyntaxError
from django.template import VariableDoesNotExist

import otree.db.models
import otree.forms

from .utils import TestCase, capture_stdout
from .models import SimplePlayer
from .simple.models import Player


class PlayerForm(otree.forms.ModelForm):
    class Meta:
        model = Player
        fields = ('my_field',)


class SimplePlayerForm(otree.forms.ModelForm):
    class Meta:
        model = SimplePlayer
        fields = ('name', 'age',)


class FormFieldTestMixin(TestCase):
    def setUp(self):
        with capture_stdout():
            call_command('create_session', 'simple', "1")
        self.player = Player.objects.first()
        self.simple_player = SimplePlayer.objects.create()

    def parse(self, fragment):
        return Template('{% load otree_tags %}' + fragment)

    def render(self, fragment, context=None):
        if context is None:
            context = Context()
        if not isinstance(context, Context):
            context = Context(context)
        return self.parse(fragment).render(context)


class FormFieldSyntaxTests(FormFieldTestMixin, TestCase):
    def test_fail_for_no_arguments(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield %}')

    def test_fail_when_multiple_fields_are_given(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field field2 %}')

    def test_fail_with_bad_arguments_after_with(self):
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field with label %}')
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field with -=- %}')
        with self.assertRaises(TemplateSyntaxError):
            self.parse('{% formfield field with label="Right" wrong %}')

    def test_one_field_is_given(self):
        self.parse('{% formfield field %}')
        self.parse('{% formfield player.field %}')
        self.parse('{% formfield player.soso.deeply.nested.field %}')

    def test_with_syntax(self):
        self.parse('{% formfield field with label="Hello" %}')
        self.parse('{% formfield field label="Hello" %}')
        self.parse('{% formfield field with help_text=variable %}')
        self.parse('''
            {% formfield field with label=label_value help_text="Constant" %}
        ''')


class FormFieldTests(FormFieldTestMixin, TestCase):
    def test_bound_field_displayed(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render('{% formfield form.name %}', {
            'form': form
        })
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''', rendered)
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name">Name:</label>
            ''', rendered)

    def test_model_field_displayed(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render('{% formfield player.name %}', {
            'form': form,
            'player': self.simple_player,
        })
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50">
            ''', rendered)
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name">Name:</label>
            ''', rendered)

    def test_prefilled_model_field_displayed(self):
        player = SimplePlayer(name='Jack')
        form = SimplePlayerForm(instance=player)
        rendered = self.render('{% formfield player.name %}', {
            'form': form,
            'player': player,
        })
        self.assertInHTML(
            '''
            <input type="text" id="id_name" name="name" class="form-control"
                maxlength="50" value="Jack">
            ''', rendered)

    def test_variable_is_none(self):
        form = SimplePlayerForm(instance=self.simple_player)
        with self.assertRaises(VariableDoesNotExist):
            self.render('{% formfield player.name %}', {
                'form': form,
                'player': None,
            })
        # Not set at all.
        with self.assertRaises(VariableDoesNotExist):
            self.render('{% formfield player.name %}', {
                'form': form,
            })

    def test_variable_is_no_model_instance(self):
        form = SimplePlayerForm(instance=self.simple_player)

        with self.assertRaises(ValueError) as cm:
            self.render('{% formfield player.name %}', {
                'form': form,
                'player': {
                    'name': 'Jack',
                }
            })

        self.assertTrue("'player.name'" in str(cm.exception))
        # The representation of the variable is also included.
        self.assertTrue("Jack" in str(cm.exception))

    def test_variable_different_to_form_instance(self):
        player1 = SimplePlayer.objects.create()
        player2 = SimplePlayer.objects.create()
        form = SimplePlayerForm(instance=player1)

        # Different model instance of the same type.
        with self.assertRaises(ValueError) as cm:
            self.render('{% formfield player.name %}', {
                'form': form,
                'player': player2,
            })
        self.assertTrue("'player'" in str(cm.exception))

    def test_model_field_not_part_of_form(self):
        form = SimplePlayerForm(instance=self.simple_player)

        # Different model instance of the same type.
        with self.assertRaises(ValueError) as cm:
            self.render('{% formfield player.unkown_field %}', {
                'form': form,
                'player': self.simple_player,
            })
        self.assertTrue("'unkown_field'" in str(cm.exception))

    def test_no_form_in_context(self):
        form = SimplePlayerForm(instance=self.simple_player)

        # Different model instance of the same type.
        with self.assertRaises(VariableDoesNotExist) as cm:
            self.render('{% formfield player.unkown_field %}', {
                'form_alternative': form,
                'player': self.simple_player,
            })

        # The exception should talk about the formfield tag and the form
        # variable.
        self.assertTrue("'formfield'" in str(cm.exception))
        self.assertTrue("'form'" in str(cm.exception))

    def test_override_label(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render(
            '{% formfield player.name with label="A fancy label" %}',
            {
                'form': form,
                'player': self.simple_player,
            })
        # Label is there.
        self.assertInHTML(
            '''
            <label for="id_name">A fancy label:</label>
            ''', rendered)

    def test_empty_label_in_bootstrap_theme(self):
        form = SimplePlayerForm(instance=self.simple_player)
        rendered = self.render(
            '''
            {% load form from floppyforms %}
            {% load formconfig from floppyforms %}
            {% form form using %}
                {% formconfig row using "floppyforms/rows/bootstrap.html" %}
                {% formfield player.name with label="" %}
            {% endform %}
            ''',
            {
                'form': form,
                'player': self.simple_player,
            })
        # Label is there.
        self.assertTrue('<label' not in rendered, rendered)



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from .models import Constants
from otree.api import WaitPage, Currency as c
from tests.utils import BlankTemplatePage as Page


class Start(Page):

    def vars_for_template(self):
        assert self.session.vars['a'] == 1
        assert self.participant.vars['a'] == 1
        assert self.participant.vars['b'] == 1
        assert self.participant.vars['c'] == Constants.complex_data_structure
        assert self.session.config['treatment'] == 'blue'

        assert self.player.in_creating_session == 1
        assert self.group.in_creating_session == 1

        return {
            'my_variable_here': 1,
        }

    def before_next_page(self):
        self.player.after_next_button_field = True
        self.session.vars['a'] = 2


class IsDisplayed(Page):

    def is_displayed(self):
        # make sure it's available during pre-fetch
        # this must be directly after a regular page
        assert self.session
        return True


class ResultsWaitPage(WaitPage):

    def after_all_players_arrive(self):
        assert self.session.vars['a'] == 2
        self.group.set_payoffs()
        for player in self.group.get_players():
            player.participant.vars['a'] = 2


class End(Page):

    def vars_for_template(self):
        assert self.player.after_next_button_field is True
        assert self.player.participant.vars['a'] == 2
        participant = self.player.participant
        assert participant.payoff == 50
        assert participant.payoff_plus_participation_fee() == 50 + 9.99
        return {}


page_sequence = [
    Start,
    IsDisplayed,
    ResultsWaitPage,
    End,
]



from __future__ import division
from . import views
from otree.api import Bot, SubmissionMustFail, Currency as c
import random


class PlayerBot(Bot):
    def play_round(self):
        yield views.Start
        yield views.IsDisplayed
        yield views.End



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants

class MyPage(Page):
    pass

page_sequence = [
        MyPage,
    ]



from .utils import TestCase
from django.core.management import call_command
from otree.models import Session
import django.test
import time
from tests.group_by_arrival_time_heartbeat.models import Subsession

class TestGBAT(TestCase):

    def setUp(self):
        call_command('create_session', 'group_by_arrival_time_heartbeat', '4')
        self.session = Session.objects.get()
        self.participants = list(self.session.get_participants())
        self.client = django.test.Client()

    def get(self, url):
        self.client.get(url, follow=True)

    def test_heartbeat(self):

        p1, p2, p3, p4 = self.participants
        self.get(p1._start_url())
        # backdate p1's visit, to simulate them being idle for a while
        p1._last_request_timestamp = time.time() - 30
        p1.save()
        self.get(p2._start_url())
        self.get(p3._start_url())

        subsession = Subsession.objects.get()
        player1, player2, player3, player4 = subsession.get_players()
        self.assertEqual(player2.group, player3.group)
        self.assertNotEqual(player1.group, player2.group)




# 2017-08-07: i temporarily disabled this feature
# If someone stores a ConstantsList into vars, it will remain a ConstantsList.
# But they need to copy it first anyway. Because what if they use it before
# it's saved and serialized?


from .utils import TestCase
from tests.constants.models import Constants
from otree.constants import MustCopyError
import random

class TestConstants(TestCase):

    def test_setattr(self):
        with self.assertRaises(AttributeError):
            Constants.c_int += 1
        with self.assertRaises(AttributeError):
            Constants.new_attr = 1

    def test_mutate_list(self):

        self.assertEqual(Constants.c_list, [1,2])

        # block direct mutation
        with self.assertRaises(MustCopyError):
            Constants.c_list[0] = 10
        with self.assertRaises(MustCopyError):
            c_list_bad = Constants.c_list
            random.shuffle(c_list_bad)
        with self.assertRaises(MustCopyError):
            Constants.c_list.append(3)

        # make sure we didn't modify it
        self.assertEqual(Constants.c_list, [1,2])

        # read-only operations are OK
        e0 = Constants.c_list[0]
        e1 = Constants.c_list[:]
        l2 = Constants.c_list + [3]
        self.assertEqual(l2, [1, 2, 3])

        # after copying, should be a normal list
        c_copied = Constants.c_list.copy()
        self.assertEqual(type(c_copied), list)
        random.shuffle(c_copied)

        c_sliced = Constants.c_list[:]
        self.assertEqual(type(c_sliced), list)
        random.shuffle(c_sliced)



from collections import namedtuple
import sys

from django.db import models
from django.template import Node
from django.template import TemplateSyntaxError
from django.template import Variable
from django.template import VariableDoesNotExist
from django.template.base import token_kwargs
from django.template.loader import get_template
from django.utils import six

from otree.models_concrete import UndefinedFormModel



class FormFieldNode(Node):
    default_template = get_template('otree/tags/_formfield.html')

    def __init__(self, field_variable_name, arg_dict):
        self.field_variable_name = field_variable_name
        self.arg_dict = arg_dict

    def get_form_instance(self, context):
        try:
            return Variable('form').resolve(context)
        except VariableDoesNotExist as exception:
            msg = (
                "The 'formfield' templatetag expects a 'form' variable "
                "in the context.")
            ExceptionClass = type(exception)
            six.reraise(
                ExceptionClass,
                ExceptionClass(msg + ' ' + str(exception)),
                sys.exc_info()[2])

    def resolve_bound_field(self, context):
        bound_field = Variable(self.field_variable_name).resolve(context)
        return bound_field

    def get_bound_field(self, context):
        # First we try to resolve the {% formfield player.name %} syntax were
        # player is a model instance.
        if '.' in self.field_variable_name:
            form = self.get_form_instance(context)
            instance_name_in_template, field_name = \
                self.field_variable_name.split('.', -1)
            instance_in_template = Variable(
                instance_name_in_template).resolve(context)
            # it could be form.my_field
            if isinstance(instance_in_template, models.Model):
                instance_from_view = form.instance
                if type(instance_from_view) == UndefinedFormModel:
                    raise ValueError(
                        'Template contains a formfield, but '
                        'you did not set form_model on the Page '
                        'in views.py.'
                    )
                elif type(instance_in_template) != type(instance_from_view):
                    raise ValueError(
                        'In views.py you set form_model to {!r}, '
                        'but in the template you have a formfield for '
                        '"{}", which is a different model.'.format(
                            type(instance_from_view),
                            instance_name_in_template,
                        )
                    )
                # we should ensure that Player and Group have readable
                # __repr__
                elif instance_in_template != instance_from_view:
                    raise ValueError(
                        "You have a formfield for '{}' "
                        "({!r}), which is different from "
                        "the expected model instance "
                        "({!r}).".format(
                            instance_name_in_template,
                            instance_in_template,
                            form.instance))
                try:
                    return form[field_name]
                except KeyError:
                    raise ValueError(
                        "'{field_name}' was used as a formfield in the template, "
                        "but was not included in the Page's 'form_fields' "
                        "in views.py ".format(
                            field_name=field_name)) from None

        # Second we try to resolve it to a bound field.
        # No field found, so we return None.
        bound_field = Variable(self.field_variable_name).resolve(context)

        # We assume it's a BoundField when 'as_widget', 'as_hidden' and
        # 'errors' attribtues are available.
        if (
                not hasattr(bound_field, 'as_widget') or
                not hasattr(bound_field, 'as_hidden') or
                not hasattr(bound_field, 'errors')):
            raise ValueError(
                "The given variable '{variable_name}' ({variable!r}) is "
                "neither a model field nor a form field.".format(
                    variable_name=self.field_variable_name,
                    variable=bound_field))
        return bound_field

    def get_extra_context(self, context):
        bound_field = self.get_bound_field(context)
        extra_context = {
            'bound_field': bound_field
        }
        if self.arg_dict:
            with_context = dict(
                (name, var.resolve(context))
                for name, var in self.arg_dict.items())
            # If the with argument label="" was set explicitly, we set it to
            # None. That is required to differentiate between 'use the default
            # label since we didn't set any in the template' and 'do not print
            # a label at all' as defined in
            # https://github.com/oTree-org/otree-core/issues/325
            if with_context.get('label', None) == '':
                with_context['label'] = None
            extra_context.update(with_context)
        return extra_context

    def render(self, context):
        extra_context = self.get_extra_context(context)
        context.update(extra_context)
        try:
            rendered = self.default_template.render(context)
        finally:
            context.pop()
        return rendered

    @classmethod
    def parse(cls, parser, token):
        bits = token.split_contents()
        tagname = bits.pop(0)
        if len(bits) < 1:
            raise TemplateSyntaxError(
                "{tagname!r} requires at least one argument.".format(
                    tagname=tagname))
        field = bits.pop(0)
        if bits:
            if bits[0] == 'with':
                bits.pop(0)
            arg_dict = token_kwargs(bits, parser, support_legacy=False)

            # Validate against spaces around the '='.
            has_lonely_equal_sign = any(
                bit.startswith('=') or bit.endswith('=')
                for bit in bits)
            if has_lonely_equal_sign:
                # If '=' is leading/trailing or is the own char in the token,
                # then the user has used spaces around it. We can use a
                # distinct error message for this to aid the user.
                raise TemplateSyntaxError(
                    "In the {tagname} tag, "
                    "you must not put spaces around the '='. "
                    "For example, do label='value', not label = 'value'."
                    ".".format(tagname=tagname))
        else:
            arg_dict = {}
        if bits:
            raise TemplateSyntaxError(
                'Unknown argument for {tagname} tag: {bits!r}'.format(
                    tagname=tagname,
                    bits=bits))
        return cls(field, arg_dict)


def defaultlabel(given_label, default):
    # We check for an explicit None here in order to allow the label to be made
    # empty.
    if given_label is None:
        return None
    if not given_label:
        return default
    return given_label



from otree.extensions import get_extensions_modules, get_extensions_data_export_views
from django.conf.urls import patterns
import inspect
from importlib import import_module

from django.conf import urls

from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.views.generic.base import RedirectView
from django.conf import settings
from django.contrib.auth.decorators import login_required
from otree.views.rest import SessionParticipantsList, Ping

from otree.common_internal import get_models_module


STUDY_UNRESTRICTED_VIEWS = {
    'AssignVisitorToRoom',
    'InitializeParticipant',
    'MTurkLandingPage',
    'MTurkStart',
    'JoinSessionAnonymously',
    'OutOfRangeNotification',
    'ParticipantRoomHeartbeat',
}


DEMO_UNRESTRICTED_VIEWS = STUDY_UNRESTRICTED_VIEWS.union({
    'AdminReport',
    'AdvanceSession',
    'CreateDemoSession',
    'DemoIndex',
    'SessionSplitScreen',
    'SessionDescription',
    'SessionMonitor',
    'SessionPayments',
    'SessionData',
    'SessionStartLinks',
    'WaitUntilSessionCreated',
})


def view_classes_from_module(module_name):
    views_module = import_module(module_name)

    # what about custom views?
    return [
        ViewCls for _, ViewCls in inspect.getmembers(views_module)
        if hasattr(ViewCls, 'url_pattern') and
        inspect.getmodule(ViewCls) == views_module
    ]


def url_patterns_from_game_module(module_name, name_in_url):
    views_module = import_module(module_name)

    all_views = [
        ViewCls
        for _, ViewCls in inspect.getmembers(views_module)
        if hasattr(ViewCls, 'url_pattern')]

    view_urls = []
    for ViewCls in all_views:

        url_pattern = ViewCls.url_pattern(name_in_url)
        url_name = ViewCls.url_name()
        view_urls.append(
            urls.url(url_pattern, ViewCls.as_view(), name=url_name)
        )

    return urls.patterns('', *view_urls)


def url_patterns_from_module(module_name):
    """automatically generates URLs for all Views in the module,
    So that you don't need to enumerate them all in urlpatterns.
    URLs take the form "gamename/ViewName".
    See the method url_pattern() for more info

    So call this function in your urls.py and pass it the names of all
    Views modules as strings.

    """

    all_views = view_classes_from_module(module_name)

    view_urls = []
    for ViewCls in all_views:
        # automatically assign URL name for reverse(), it defaults to the
        # class's name
        url_name = getattr(ViewCls, 'url_name', ViewCls.__name__)

        if settings.AUTH_LEVEL == 'STUDY':
            unrestricted = url_name in STUDY_UNRESTRICTED_VIEWS
        elif settings.AUTH_LEVEL == 'DEMO':
            unrestricted = url_name in DEMO_UNRESTRICTED_VIEWS
        else:
            unrestricted = True

        if unrestricted:
            as_view = ViewCls.as_view()
        else:
            as_view = login_required(ViewCls.as_view())

        url_pattern = ViewCls.url_pattern
        if callable(url_pattern):
            url_pattern = url_pattern()

        view_urls.append(
            urls.url(url_pattern, as_view, name=url_name)
        )

    return urls.patterns('', *view_urls)


def extensions_urlpatterns():

    urlpatterns = []

    for url_module in get_extensions_modules('urls'):
        urlpatterns += getattr(url_module, 'urlpatterns', [])

    return urlpatterns


def extensions_export_urlpatterns():
    view_classes = get_extensions_data_export_views()
    view_urls = []

    for ViewCls in view_classes:
        if settings.AUTH_LEVEL in {'DEMO', 'STUDY'}:
            as_view = login_required(ViewCls.as_view())
        else:
            as_view = ViewCls.as_view()
        view_urls.append(urls.url(ViewCls.url_pattern, as_view, name=ViewCls.url_name))

    return urls.patterns('', *view_urls)


def augment_urlpatterns(urlpatterns):

    urlpatterns += urls.patterns(
        '',
        urls.url(r'^$', RedirectView.as_view(url='/demo', permanent=True)),
        urls.url(
            r'^accounts/login/$',
            'django.contrib.auth.views.login',
            {'template_name': 'otree/login.html'},
            name='login_url',
        ),
        urls.url(
            r'^accounts/logout/$',
            'django.contrib.auth.views.logout',
            {'next_page': 'DemoIndex'},
            name='logout',
        ),
    )

    rest_api_urlpatterns = (
        urls.url(r'^ping/$', Ping.as_view(), name="ping"),
        urls.url(
            r'^sessions/(?P<session_code>[a-z0-9]+)/participants/$',
            SessionParticipantsList.as_view(),
            name="session_participants_list")
    )
    urlpatterns += rest_api_urlpatterns

    urlpatterns += staticfiles_urlpatterns()

    used_names_in_url = set()
    for app_name in settings.INSTALLED_OTREE_APPS:
        models_module = get_models_module(app_name)
        name_in_url = models_module.Constants.name_in_url
        if name_in_url in used_names_in_url:
            msg = (
                "App {} has name_in_url='{}', "
                "which is already used by another app"
            ).format(app_name, name_in_url)
            raise ValueError(msg)

        used_names_in_url.add(name_in_url)
        views_module_name = '{}.views'.format(app_name)
        urlpatterns += url_patterns_from_game_module(
            views_module_name, name_in_url)

    urlpatterns += url_patterns_from_module('otree.views.concrete')
    urlpatterns += url_patterns_from_module('otree.views.demo')
    urlpatterns += url_patterns_from_module('otree.views.admin')
    urlpatterns += url_patterns_from_module('otree.views.mturk')
    urlpatterns += url_patterns_from_module('otree.views.export')

    urlpatterns += extensions_urlpatterns()
    urlpatterns += extensions_export_urlpatterns()

    return urlpatterns


urlpatterns = patterns('',)
augment_urlpatterns(urlpatterns)



from collections import namedtuple
from django.core.management import call_command
from django.core.urlresolvers import reverse
from otree.common import RealWorldCurrency
from otree.models import Session, Participant
from otree.views.mturk import MTurkClient
import django.test.client
from unittest import mock
from unittest.mock import MagicMock
from .utils import TestCase
from django.test import override_settings, LiveServerTestCase
from tests.utils import OTreePhantomBrowser

from otree.views.mturk import get_mturk_client


class TestMTurk(TestCase):

    def setUp(self):
        call_command('create_session', 'mturk', "2")
        self.session = Session.objects.get()
        self.browser = django.test.client.Client()

    def test_get_create_hit(self):
        url = reverse('MTurkCreateHIT', args=(self.session.code,))
        response = self.browser.get(url, follow=True)
        self.assertEqual(response.status_code, 200)

    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_post_create_hit(self, mock_get_mturk_client):
        mocked_client = mock_get_mturk_client()

        mocked_client.create_hit.return_value = {'HIT': {'HITId': 'AAA', 'HITGroupId': 'BBB'}}
        url = reverse('MTurkCreateHIT', args=(self.session.code,))

        participation_fee = 1.47
        assignments = 6
        minutes_allotted_per_assignment = 60
        expiration_hours = 168

        response = self.browser.post(
            url,
            data=dict(
                in_sandbox = True,
                title = 'Title for your experiment',
                description = 'Description for your experiment',
                keywords = 'easy, bonus, choice, study',
                # different from the session config's default
                money_reward = participation_fee,
                assignments = assignments,
                minutes_allotted_per_assignment = minutes_allotted_per_assignment,
                expiration_hours = expiration_hours,
            ),
            follow=True
        )

        self.assertEqual(response.status_code, 200)

        mocked_client.create_hit.assert_called()
        # just a subset, things that are easy to mock
        expected_call_dict_subset = {
            'Title': 'Title for your experiment',
            'Description': 'Description for your experiment',
            'Keywords': 'easy, bonus, choice, study',
            'MaxAssignments': assignments,
            'Reward': str(participation_fee),
            'AssignmentDurationInSeconds': minutes_allotted_per_assignment*60,
            'LifetimeInSeconds': 60*60*expiration_hours
        }
        actual_call_dict = mocked_client.create_hit.call_args[1]
        self.assertLessEqual(expected_call_dict_subset.items(), actual_call_dict.items())

        mocked_client.create_hit.assert_called()

        session = Session.objects.get()
        self.assertEqual(session.config['participation_fee'], 1.47)


    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_mturk_start(self, mock_get_mturk_client):
        mocked_client = mock_get_mturk_client.return_value

        url = reverse('MTurkStart', args=(self.session.code,))
        # needs to start with 'A'
        worker_id = 'AWORKERID1'
        assignment_id = 'ASSIGNMENTID1'
        response = self.browser.get(
            url,
            data={
                'assignmentId': assignment_id,
                'workerId': worker_id
            },
            follow=True
        )
        self.assertEqual(response.status_code, 200)

        # worker returns the assignment, accepts another assignment
        assignment_id2 = 'ASSIGNMENTID2'
        response = self.browser.get(
            url,
            data={
                'assignmentId': assignment_id2,
                'workerId': worker_id
            },
            follow=True
        )

        mocked_client.associate_qualification_with_worker.assert_called()

        # should be only 1
        p_visitor = Participant.objects.get(
            session=self.session,
            visited=True
        )
        p_non_visitor = Participant.objects.filter(
            session=self.session,
            visited=False
        )[0]

        self.assertEqual(p_visitor.mturk_worker_id, worker_id)
        self.assertEqual(p_visitor.mturk_assignment_id, assignment_id2)
        self.assertEqual(p_non_visitor.mturk_worker_id, None)
        self.assertEqual(p_non_visitor.mturk_assignment_id, None)

    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_mturk_preview(self, mock_get_mturk_client):
        url = reverse('MTurkLandingPage', args=(self.session.code,))
        assignment_id = 'ASSIGNMENT_ID01'
        worker_id = 'WORKER_ID01'
        response = self.browser.get(
            url,
            data={
                'assignmentId': assignment_id,
                'workerId': worker_id
            },
            follow=True
        )
        self.assertEqual(response.status_code, 200)



class TestPayMTurk(LiveServerTestCase):

    '''
    Use LiveServerTestCase instead of ChannelLiveServerTestCase, because
    I need to use mocks, which are not possible with ChannelLiveServerTestCase
    because it uses a separate process
    '''

    def setUp(self):
        num_participants = 6
        # create a session
        call_command('create_session', 'mturk', str(num_participants))
        self.session = Session.objects.get()
        self.session.mturk_HITId = 'FAKEHITID'
        self.session.save()
        self.browser = OTreePhantomBrowser(live_server_url=self.live_server_url)
        self.participants = list(self.session.get_participants())

    # so we don't have to worry about points conversion.
    @override_settings(USE_POINTS=False)
    @mock.patch('otree.views.mturk.get_mturk_client')
    def test_pay(self, mock_get_mturk_client):

        mocked_client = mock_get_mturk_client()

        bonus_per_participant = 0.2
        self.assertEqual(self.session.config['participation_fee'], 0.4)
        total_pay_per_participant = 0.6

        # simulate workers arriving from MTurk by assigning bogus worker IDs
        # and assignment IDs
        for p in self.participants:
            p.mturk_assignment_id = p.mturk_worker_id = str(p.id_in_session)
            p.payoff = bonus_per_participant
            p.save()

        # mock the result set, so that when the view queries MTurk to see
        # who is submitted, it returns a list of MTurk workers

        def list_assignments_for_hit(**kwargs):
            # don't paginate
            if 'NextToken' in kwargs:
                return {'Assignments': []}
            return {
                'NextToken': 'foo',
                'Assignments': [
                    {'WorkerId': p.mturk_worker_id, 'AssignmentStatus': 'Submitted'}
                    # simulate that not everyone submitted
                    for p in self.participants[:-2]
                ]
            }

        mocked_client.list_assignments_for_hit = list_assignments_for_hit

        url = reverse('MTurkSessionPayments', args=[self.session.code])
        br = self.browser
        br.go(url)

        br.find_by_css('[name="workers"][value="1"]').check()
        br.find_by_css('[name="workers"][value="2"]').check()

        button = br.find_by_css('button#pay')
        self.assertEqual(button.text, 'Pay via MTurk ($1.20)')
        button.click()

        confirm_button = br.find_by_id('pay-confirm-button')
        confirm_button.click()

        calls = mocked_client.approve_assignment.call_args_list
        for asst_id in ['1', '2']:
            mocked_client.approve_assignment.assert_any_call(
                AssignmentId=asst_id)

        calls = mocked_client.send_bonus.call_args_list
        self.assertEqual(len(calls), 2)

        for call in calls:
            kwargs = call[1]
            self.assertEqual(kwargs['BonusAmount'], '0.20')
            # because our bogus worker ID is the same as our bogus assignment ID
            self.assertEqual(kwargs['WorkerId'], kwargs['AssignmentId'])

        # after submitting, oTree redirects to the payments page
        br.find_by_css('[name="workers"][value="3"]').check()
        br.find_by_css('[name="workers"][value="4"]').check()

        button = br.find_by_css('button#reject')
        button.click()
        reject_confirm_button = br.find_by_id('reject-confirm-button')
        reject_confirm_button.click()

        calls = mocked_client.reject_assignment.call_args_list
        self.assertEqual(len(calls), 2)
        for asst_id in ['3', '4']:
            mocked_client.reject_assignment.assert_any_call(
                AssignmentId=asst_id, RequesterFeedback='')



from channels.routing import route, route_class

from otree.channels import consumers
from otree.extensions import get_extensions_modules


channel_routing = [

    route('otree.create_session',
          consumers.create_session),

    # WebSockets
    route_class(
        consumers.WaitPage,
        path=r'^/wait_page/(?P<params>[\w,]+)/$'),
    route_class(
        consumers.GroupByArrivalTime,
        path=r'^/group_by_arrival_time/(?P<params>[\w,\.]+)/$'),
    route_class(
        consumers.AutoAdvance,
        path=r'^/auto_advance/(?P<params>[\w,]+)/$'),
    route_class(consumers.WaitForSession,
          path=r'^/wait_for_session/(?P<pre_create_id>\w+)/$'),
    route_class(
          consumers.RoomParticipant,
          path=r'^/wait_for_session_in_room/(?P<params>[\w,]+)/$'),
    route_class(
          consumers.RoomAdmin,
          path=r'^/room_without_session/(?P<room>\w+)/$'),
    route_class(
          consumers.BrowserBotsClient,
          path=r'^/browser_bots_client/(?P<session_code>\w+)/$'),
    route_class(
          consumers.BrowserBot,
          path=r'^/browser_bot_wait/$'),


]

for extensions_module in get_extensions_modules('routing'):
    channel_routing += getattr(extensions_module, 'channel_routing', [])



import time
import threading
import django.utils.timezone
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404, render_to_response
from django.template.response import TemplateResponse
from django.http import (
    HttpResponse, HttpResponseRedirect,
    HttpResponseNotFound, Http404, JsonResponse
)
from django.utils.translation import ugettext as _

import vanilla

import otree.constants_internal as constants
import otree.models
from otree.models import Participant, Session
from otree.common_internal import (
    make_hash, add_params_to_url, get_redis_conn
)
import otree.views.admin
import otree.views.mturk
import otree.common_internal
from otree.views.abstract import (
    GenericWaitPageMixin,
    get_redis_lock, NO_PARTICIPANTS_LEFT_MSG)
from otree.room import ROOM_DICT
from otree.models_concrete import (
    ParticipantRoomVisit, BrowserBotsLauncherSessionCode)

start_link_thread_lock = threading.RLock()

class OutOfRangeNotification(vanilla.View):
    name_in_url = 'shared'

    def dispatch(self, request, *args, **kwargs):
        return TemplateResponse(
            request, 'otree/OutOfRangeNotification.html'
        )

    url_pattern = '^OutOfRangeNotification/$'


class InitializeParticipant(vanilla.UpdateView):
    """just collects data and sets properties. not essential to functionality.
    the only exception is if the participant needs to be assigned to groups on
    the fly, which is done here.

    2014-11-16: also, this sets _last_page_timestamp. what if that is not set?
    will it still work?

    """

    url_pattern = r'^InitializeParticipant/(?P<{}>[a-z0-9]+)/$'.format(
            constants.participant_code
        )

    def get(self, *args, **kwargs):

        participant = get_object_or_404(
            Participant,
            code=kwargs[constants.participant_code]
        )

        if participant._index_in_pages == 0:
            participant._index_in_pages = 1
            participant.visited = True

            # participant.label might already have been set
            participant.label = participant.label or self.request.GET.get(
                constants.participant_label
            )
            participant.ip_address = self.request.META['REMOTE_ADDR']

            now = django.utils.timezone.now()
            participant.time_started = now
            participant._last_page_timestamp = time.time()

            participant.save()
        first_url = participant._url_i_should_be_on()
        return HttpResponseRedirect(first_url)


class MTurkLandingPage(vanilla.TemplateView):

    def get_template_names(self):
        hit_settings = self.session.config['mturk_hit_settings']
        return [hit_settings['preview_template']]

    url_pattern = r"^MTurkLandingPage/(?P<session_code>[a-z0-9]+)/$"

    def dispatch(self, request, *args, **kwargs):
        session_code = kwargs['session_code']
        self.session = get_object_or_404(
            otree.models.Session, code=session_code
        )
        return super().dispatch(
            request, *args, **kwargs
        )

    def get(self, request, *args, **kwargs):
        assignment_id = (
            self.request.GET['assignmentId']
            if 'assignmentId' in self.request.GET else
            ''
        )
        if assignment_id and assignment_id != 'ASSIGNMENT_ID_NOT_AVAILABLE':
            url_start = reverse('MTurkStart', args=(self.session.code,))
            url_start = add_params_to_url(url_start, {
                'assignmentId': self.request.GET['assignmentId'],
                'workerId': self.request.GET['workerId']})
            return HttpResponseRedirect(url_start)

        context = super().get_context_data(**kwargs)
        return self.render_to_response(context)


class MTurkStart(vanilla.View):

    url_pattern = r"^MTurkStart/(?P<session_code>[a-z0-9]+)/$"

    def dispatch(self, request, *args, **kwargs):
        session_code = kwargs['session_code']
        self.session = get_object_or_404(
            otree.models.Session, code=session_code
        )
        return super(MTurkStart, self).dispatch(
            request, *args, **kwargs
        )

    def get(self, *args, **kwargs):
        assignment_id = self.request.GET['assignmentId']
        worker_id = self.request.GET['workerId']
        qualification_id = self.session.config['mturk_hit_settings'].get('grant_qualification_id')
        if qualification_id:
            # don't pass request arg, because we don't want to show a message.
            # using the fully qualified name because that seems to make mock.patch work
            mturk_client = otree.views.mturk.get_mturk_client(
                use_sandbox=self.session.mturk_use_sandbox)
            # seems OK to assign this multiple times
            mturk_client.associate_qualification_with_worker(
                QualificationTypeId=qualification_id,
                WorkerId=worker_id,
                # Mturk complains if I omit IntegerValue
                IntegerValue=1
            )
        try:
            # just check if this worker already game, but
            # don't filter for assignment, because maybe they already started
            # and returned the previous assignment
            # in this case, we should assign back to the same participant
            # so that we don't get duplicates in the DB, and so people
            # can't snoop and try the HIT first, then re-try to get a bigger bonus
            participant = self.session.participant_set.get(
                mturk_worker_id=worker_id)
        except Participant.DoesNotExist:
            with get_redis_lock(name='start_links') or start_link_thread_lock:
                try:
                    participant = self.session.participant_set.filter(
                        visited=False
                    ).order_by('start_order')[0]
                except IndexError:
                    return HttpResponseNotFound(NO_PARTICIPANTS_LEFT_MSG)

                # 2014-10-17: needs to be here even if it's also set in
                # the next view to prevent race conditions
                # this needs to be inside the lock
                participant.visited = True
                participant.mturk_worker_id = worker_id
        # reassign assignment_id, even if they are returning, because maybe they accepted
        # and then returned, then re-accepted with a different assignment ID
        # if it's their second time
        participant.mturk_assignment_id = assignment_id
        participant.save()
        return HttpResponseRedirect(participant._start_url())


def get_existing_or_new_participant(session, label):
    if label:
        try:
            return session.participant_set.get(label=label)
        except Participant.DoesNotExist:
            pass
    return session.participant_set.filter(
        visited=False).order_by('start_order').first()


def get_participant_with_cookie_check(session, cookies):
    cookie_name = 'session_{}_participant'.format(session.code)
    participant_code = cookies.get(cookie_name)
    # this could return None
    if participant_code:
        return Participant.objects.filter(code=participant_code).first()
    participant = session.participant_set.filter(
        visited=False).order_by('start_order').first()
    if participant:
        cookies[cookie_name] = participant.code
        return participant


def participant_start_page_or_404(session, *, label, cookies=None):
    '''pass request.session as an arg if you want to get/set a cookie'''
    with get_redis_lock(name='start_links') or start_link_thread_lock:
        if cookies is None:
            participant = get_existing_or_new_participant(session, label)
        else:
            participant = get_participant_with_cookie_check(session, cookies)
        if not participant:
            return HttpResponseNotFound(NO_PARTICIPANTS_LEFT_MSG)

        # needs to be here even if it's also set in
        # the next view to prevent race conditions
        participant.visited = True
        if label:
            participant.label = label
        participant.save()

    return HttpResponseRedirect(participant._start_url())


class JoinSessionAnonymously(vanilla.View):

    url_pattern = r'^join/(?P<anonymous_code>[a-z0-9]+)/$'

    def get(self, *args, **kwargs):

        anonymous_code = kwargs['anonymous_code']
        session = get_object_or_404(
            otree.models.Session, _anonymous_code=anonymous_code
        )
        label = self.request.GET.get('participant_label')
        return participant_start_page_or_404(session, label=label)


class AssignVisitorToRoom(GenericWaitPageMixin, vanilla.View):

    url_pattern = r'^room/(?P<room>\w+)/$'

    def dispatch(self, request, *args, **kwargs):
        self.room_name = kwargs['room']
        try:
            room = ROOM_DICT[self.room_name]
        except KeyError:
            return HttpResponseNotFound('Invalid room specified in url')

        self.uses_pin = room.has_pin_code()

        label = self.request.GET.get(
            'participant_label', ''
        )

        if room.has_participant_labels():
            if not label:
                if not room.use_secure_urls:
                    return render_to_response("otree/RoomInputLabel.html")

            if not label in room.get_participant_labels():
                return HttpResponseNotFound(
                    _('Invalid participant label.')
                )

            if room.use_secure_urls:
                hash = self.request.GET.get('hash')
                if hash != make_hash(label):
                    return HttpResponseNotFound('Invalid hash parameter.')

        if self.uses_pin:
            pin_code = self.request.GET.get('pin')
            if not pin_code:
                return super(AssignVisitorToRoom, self).get(args, kwargs)

            if pin_code != room.get_pin_code():
                return HttpResponseNotFound('The given pin code is incorrect.')

        session = room.session
        if session is None:
            self.tab_unique_id = otree.common_internal.random_chars_10()
            self._socket_url_params = ','.join([
                self.room_name,
                label,
                # random chars in case the participant has multiple tabs open
                self.tab_unique_id,
            ])
            return render_to_response(
                "otree/WaitPageRoom.html",
                {
                    'view': self, 'title_text': _('Please wait'),
                    'body_text': _('Waiting for your session to begin')
                }
            )

        if label:
            cookies = None
        else:
            cookies = request.session


        # 2017-08-02: changing the behavior so that even in a room without
        # participant_label_file, 2 requests for the same start URL with same label
        # will return the same participant. Not sure if the previous behavior
        # (assigning to 2 different participants) was intentional or bug.
        return participant_start_page_or_404(session, label=label, cookies=cookies)

    def get_context_data(self, **kwargs):
        return {
            'room': self.room_name,
            'uses_pin': self.uses_pin,
        }

    def socket_url(self):
        return '/wait_for_session_in_room/{}/'.format(
            self._socket_url_params
        )

    def redirect_url(self):
        return self.request.get_full_path()


class StaleRoomVisits(vanilla.View):

    url_pattern = r'^StaleRoomVisits/(?P<room>\w+)/$'

    def get(self, request, *args, **kwargs):
        stale_threshold = time.time() - 20
        stale_participant_labels = ParticipantRoomVisit.objects.filter(
            room_name=kwargs['room'],
            last_updated__lt=stale_threshold
        ).values_list('participant_label', flat=True)

        # make json serializable
        stale_participant_labels = list(stale_participant_labels)

        return JsonResponse({'participant_labels': stale_participant_labels})


class ActiveRoomParticipantsCount(vanilla.View):

    url_pattern = r'^ActiveRoomParticipantsCount/(?P<room>\w+)/$'

    def get(self, request, *args, **kwargs):
        count = ParticipantRoomVisit.objects.filter(
            room_name=kwargs['room'],
            last_updated__gte=time.time() - 20
        ).count()

        return JsonResponse({'count': count})


class ParticipantRoomHeartbeat(vanilla.View):

    url_pattern = r'^ParticipantRoomHeartbeat/(?P<tab_unique_id>\w+)/$'

    def get(self, request, *args, **kwargs):
        # better not to return 404, because in practice, on Firefox,
        # this was still being requested after the session started.
        ParticipantRoomVisit.objects.filter(
            tab_unique_id=kwargs['tab_unique_id']
        ).update(last_updated=time.time())
        return HttpResponse('')


class ParticipantHeartbeatGBAT(vanilla.View):
    url_pattern = r'^ParticipantHeartbeatGBAT/(?P<participant_code>\w+)/$'

    def get(self, request, *args, **kwargs):
        Participant.objects.filter(code=kwargs['participant_code']).update(
            _last_request_timestamp=time.time())
        return HttpResponse('')


class AdvanceSession(vanilla.View):

    url_pattern = r'^AdvanceSession/(?P<session_code>[a-z0-9]+)/$'

    def post(self, *args, **kwargs):
        session = get_object_or_404(
            otree.models.Session, code=kwargs['session_code']
        )
        session.advance_last_place_participants()
        return HttpResponse('ok')


class ToggleArchivedSessions(vanilla.View):

    url_pattern = r'^ToggleArchivedSessions/'

    def post(self, request, *args, **kwargs):
        code_list = request.POST.getlist('item-action')
        sessions = otree.models.Session.objects.filter(
            code__in=code_list)
        code_dict = {True: [], False: []}
        for code, archived in sessions.values_list('code', 'archived'):
            code_dict[archived].append(code)

        for code in code_list:
            if not (code in code_dict[True] or code in code_dict[False]):
                raise Http404('No session with the code %s.' % code)

        # TODO: When `F` implements a toggle, use this instead:
        #       sessions.update(archived=~F('archived'))
        otree.models.Session.objects.filter(
            code__in=code_dict[True]).update(archived=False)
        otree.models.Session.objects.filter(
            code__in=code_dict[False]).update(archived=True)

        return HttpResponseRedirect(request.POST['origin_url'])


class DeleteSessions(vanilla.View):

    url_pattern = r'^DeleteSessions/'

    def dispatch(self, *args, **kwargs):
        return super(DeleteSessions, self).dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        for code in request.POST.getlist('item-action'):
            session = get_object_or_404(
                otree.models.Session, code=code
            )
            session.delete()
        return HttpResponseRedirect(reverse('Sessions'))


class BrowserBotStartLink(GenericWaitPageMixin, vanilla.View):

    url_pattern = r'^browser_bot_start/$'

    def dispatch(self, request, *args, **kwargs):
        get_redis_conn()
        session_info = BrowserBotsLauncherSessionCode.objects.first()
        if session_info:
            session = Session.objects.get(code=session_info.code)
            with get_redis_lock(name='start_links') or start_link_thread_lock:
                participant = session.participant_set.filter(
                    visited=False).order_by('start_order').first()
                if not participant:
                    return HttpResponseNotFound(NO_PARTICIPANTS_LEFT_MSG)

                # 2014-10-17: needs to be here even if it's also set in
                # the next view to prevent race conditions
                participant.visited = True
                participant.save()

            return HttpResponseRedirect(participant._start_url())
        else:
            ctx = {'view': self, 'title_text': 'Please wait',
                   'body_text': 'Waiting for browser bots session to begin'}
            return render_to_response("otree/WaitPage.html", ctx)

    def socket_url(self):
        return '/browser_bot_wait/'

    def redirect_url(self):
        return self.request.get_full_path()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import django.db
import django.utils.timezone
import traceback
import time

from channels import Group
from channels.generic.websockets import JsonWebsocketConsumer

import otree.session
from otree.models import Participant, Session
from otree.models_concrete import (
    CompletedGroupWaitPage, CompletedSubsessionWaitPage)
from otree.common_internal import (
    channels_wait_page_group_name, channels_create_session_group_name,
    channels_group_by_arrival_time_group_name, get_models_module,
    channels_room_participants_group_name
)
from otree.models_concrete import (
    FailedSessionCreation, ParticipantRoomVisit,
    FAILURE_MESSAGE_MAX_LENGTH, BrowserBotsLauncherSessionCode)
from otree.room import ROOM_DICT

logger = logging.getLogger(__name__)


class OTreeJsonWebsocketConsumer(JsonWebsocketConsumer):
    '''
    THIS IS NOT PUBLIC API.
    Third party apps should not subclass this.
    Either copy this class into your code,
    or subclass directly from JsonWebsocketConsumer,
    '''

    def clean_kwargs(self, **kwargs):
        '''
        subclasses should override if the route receives a comma-separated params arg.
        otherwise, this just passes the route kwargs as is (usually there is just one).
        The output of this method is passed to self.group_name(), self.post_connect,
        and self.pre_disconnect, so within each class, all 3 of those methods must
        accept the same args (or at least take a **kwargs wildcard, if the args aren't used)
        '''
        return kwargs

    def group_name(self, **kwargs):
        raise NotImplementedError()

    def connection_groups(self, **kwargs):
        kwargs = self.clean_kwargs(**kwargs)
        group_name = self.group_name(**kwargs)
        return [group_name]

    def connect(self, message, **kwargs):
        # don't do this until we upgrade to channels 1.0+
        # self.message.reply_channel.send({"accept": True})
        kwargs = self.clean_kwargs(**kwargs)
        self.post_connect(**kwargs)

    def post_connect(self, **kwargs):
        pass

    def disconnect(self, message, **kwargs):
        kwargs = self.clean_kwargs(**kwargs)
        self.pre_disconnect(**kwargs)

    def pre_disconnect(self, **kwargs):
        pass


class GroupByArrivalTime(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        session_pk, page_index, app_name, player_id = params.split(',')
        return {
            'app_name': app_name,
            'session_pk': int(session_pk),
            'page_index': int(page_index),
            'player_id': int(player_id)
        }

    def group_name(self, app_name, player_id, page_index, session_pk):
        return channels_group_by_arrival_time_group_name(
            session_pk, page_index)

    def post_connect(self, app_name, player_id, page_index, session_pk):
        models_module = get_models_module(app_name)
        group_id_in_subsession = models_module.Group.objects.filter(
            player__id=player_id).values_list(
            'id_in_subsession', flat=True)[0]

        ready = CompletedGroupWaitPage.objects.filter(
            page_index=page_index,
            id_in_subsession=int(group_id_in_subsession),
            session_id=session_pk,
            ).exists()
        if ready:
            self.send({'status': 'ready'})


class WaitPage(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        session_pk, page_index, group_id_in_subsession = params.split(',')
        return {
            'session_pk': int(session_pk),
            'page_index': int(page_index),
            # don't convert group_id_in_subsession to int yet, it might be null
            'group_id_in_subsession': group_id_in_subsession,
        }

    def group_name(self, session_pk, page_index, group_id_in_subsession):
        return channels_wait_page_group_name(
                session_pk, page_index, group_id_in_subsession)

    def post_connect(self, session_pk, page_index, group_id_in_subsession):
        # in case message was sent before this web socket connects
        if group_id_in_subsession:
            ready = CompletedGroupWaitPage.objects.filter(
                page_index=page_index,
                id_in_subsession=int(group_id_in_subsession),
                session_id=session_pk,
                ).exists()
        else:  # subsession
            ready = CompletedSubsessionWaitPage.objects.filter(
                page_index=page_index,
                session_id=session_pk,
                ).exists()
        if ready:
            self.send({'status': 'ready'})


class AutoAdvance(OTreeJsonWebsocketConsumer):
    def clean_kwargs(self, params):
        participant_code, page_index = params.split(',')
        return {
            'participant_code': participant_code,
            'page_index': int(page_index),
        }

    def group_name(self, page_index, participant_code):
        return 'auto-advance-{}'.format(participant_code)

    def post_connect(self, page_index, participant_code):
        # in case message was sent before this web socket connects
        result = Participant.objects.filter(
                code=participant_code).values_list(
            '_index_in_pages', flat=True)
        try:
            page_should_be_on = result[0]
        except IndexError:
            # doesn't get shown because not yet localized
            self.send({'error': 'Participant not found in database.'})
            return
        if page_should_be_on > page_index:
            self.send({'auto_advanced': True})


def create_session(message):
    group = Group(message['channels_group_name'])

    kwargs = message['kwargs']

    # because it's launched through web UI
    kwargs['honor_browser_bots_config'] = True
    try:
        otree.session.create_session(**kwargs)
    except Exception as e:

        # full error message is printed to console (though sometimes not?)
        error_message = 'Failed to create session: "{}"'.format(e)
        traceback_str = traceback.format_exc()
        group.send(
            {'text': json.dumps(
                {
                    'error': error_message,
                    'traceback': traceback_str,
                })}
        )
        FailedSessionCreation.objects.create(
            pre_create_id=kwargs['_pre_create_id'],
            message=error_message[:FAILURE_MESSAGE_MAX_LENGTH],
            traceback=traceback_str
        )
        raise

    group.send(
        {'text': json.dumps(
            {'status': 'ready'})}
    )

    if 'room_name' in kwargs:
        Group(channels_room_participants_group_name(kwargs['room_name'])).send(
            {'text': json.dumps(
                {'status': 'session_ready'})}
        )


class WaitForSession(OTreeJsonWebsocketConsumer):
    def clean_kwargs(self, **kwargs):
        return kwargs

    def group_name(self, pre_create_id):
        return channels_create_session_group_name(pre_create_id)

    def post_connect(self, pre_create_id):

        group_name = self.group_name(pre_create_id)

        # in case message was sent before this web socket connects
        if Session.objects.filter(
                _pre_create_id=pre_create_id, ready=True).exists():
            self.group_send(group_name, {'status': 'ready'})
        else:
            failure = FailedSessionCreation.objects.filter(
                pre_create_id=pre_create_id
            ).first()
            if failure:
                self.group_send(group_name,
                        {'error': failure.message,
                         'traceback': failure.traceback})


class RoomAdmin(OTreeJsonWebsocketConsumer):

    def group_name(self, room):
        return 'room-admin-{}'.format(room)

    def post_connect(self, room):

        room_object = ROOM_DICT[room]

        now = time.time()
        stale_threshold = now - 15
        present_list = ParticipantRoomVisit.objects.filter(
            room_name=room_object.name,
            last_updated__gte=stale_threshold,
        ).values_list('participant_label', flat=True)

        # make it JSON serializable
        present_list = list(present_list)

        self.send({
            'status': 'load_participant_lists',
            'participants_present': present_list,
        })

        # prune very old visits -- don't want a resource leak
        # because sometimes not getting deleted on WebSocket disconnect
        very_stale_threshold = now - 10*60
        ParticipantRoomVisit.objects.filter(
            room_name=room_object.name,
            last_updated__lt=very_stale_threshold,
        ).delete()


class RoomParticipant(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        room_name, participant_label, tab_unique_id = params.split(',')
        return {
            'room_name': room_name,
            'participant_label': participant_label,
            'tab_unique_id': tab_unique_id,
        }

    def group_name(self, room_name, participant_label, tab_unique_id):
        return channels_room_participants_group_name(room_name)

    def post_connect(self, room_name, participant_label, tab_unique_id):
        if room_name in ROOM_DICT:
            room = ROOM_DICT[room_name]
        else:
            # doesn't get shown because not yet localized
            self.send({'error': 'Invalid room name "{}".'.format(room_name)})
            return
        if room.has_session():
            self.send({'status': 'session_ready'})
        else:
            try:
                ParticipantRoomVisit.objects.create(
                    participant_label=participant_label,
                    room_name=room_name,
                    tab_unique_id=tab_unique_id,
                    last_updated=time.time(),
                )
            except django.db.IntegrityError as exc:
                # possible that the tab connected twice
                # without disconnecting in between
                # because of WebSocket failure
                # tab_unique_id is unique=True,
                # so this will throw an integrity error.
                logger.info(
                    'ParticipantRoomVisit: not creating a new record because a '
                    'database integrity error was thrown. '
                    'The exception was: {}: {}'.format(type(exc), exc))
                pass
            self.group_send(
                'room-admin-{}'.format(room_name),
                {
                    'status': 'add_participant',
                    'participant': participant_label
                }
            )

    def pre_disconnect(self, room_name, participant_label, tab_unique_id):

        if room_name in ROOM_DICT:
            room = ROOM_DICT[room_name]
        else:
            # doesn't get shown because not yet localized
            self.send({'error': 'Invalid room name "{}".'.format(room_name)})
            return

        # should use filter instead of get,
        # because if the DB is recreated,
        # the record could already be deleted
        ParticipantRoomVisit.objects.filter(
            participant_label=participant_label,
            room_name=room_name,
            tab_unique_id=tab_unique_id).delete()

        if room.has_participant_labels():
            if not ParticipantRoomVisit.objects.filter(
                participant_label=participant_label,
                room_name=room_name
            ).exists():
                # it's ok if there is a race condition --
                # in JS removing a participant is idempotent
                self.group_send(
                    'room-admin-{}'.format(room_name),
                    {
                        'status': 'remove_participant',
                        'participant': participant_label
                    }
                )
        else:
            self.group_send(
                'room-admin-{}'.format(room_name),
                {
                    'status': 'remove_participant',
                }
            )


class BrowserBotsClient(OTreeJsonWebsocketConsumer):

    def group_name(self, session_code):
        return 'browser-bots-client-{}'.format(self.kwargs['session_code'])


class BrowserBot(OTreeJsonWebsocketConsumer):

    def group_name(self):
        return 'browser_bot_wait'

    def post_connect(self):
        launcher_session_info = BrowserBotsLauncherSessionCode.objects.first()
        if launcher_session_info:
            self.send({'status': 'session_ready'})



"""oTree Public API utilities"""

import json
from decimal import Decimal

from django.conf import settings
from django.utils import formats, numberformat
from django.utils.safestring import mark_safe
from django.utils.translation import ungettext

import six

import easymoney


# =============================================================================
# MONKEY PATCH - fix for https://github.com/oTree-org/otree-core/issues/387
# =============================================================================

# Black Magic: The original number format of django used inside templates don't
# work if the currency code contains non-ascii characters. This ugly hack
# remplace the original number format and when you has a easy_money instance
# simple use the old unicode casting.

_original_number_format = numberformat.format


def otree_number_format(number, *args, **kwargs):
    if isinstance(number, easymoney.Money):
        return six.text_type(number)
    return _original_number_format(number, *args, **kwargs)

numberformat.format = otree_number_format


# =============================================================================
# CLASSES
# =============================================================================

class RealWorldCurrency(easymoney.Money):
    '''payment currency'''

    CODE = settings.REAL_WORLD_CURRENCY_CODE
    LOCALE = settings.REAL_WORLD_CURRENCY_LOCALE
    DECIMAL_PLACES = settings.REAL_WORLD_CURRENCY_DECIMAL_PLACES

    def to_number(self):
        '''DEPRECATED. Don't use this.'''
        return Decimal(self)



class Currency(RealWorldCurrency):
    '''game currency'''

    # if I uncomment the below,
    # it should solve the problem in PyCharm where currency amounts are
    # highlighted in yellow if you try to divide
    # by a number. but there are other currency problems, like adding
    # to a number. Or even adding IntegerField to a number.
    # so i want to understand better first how to solve the problem generally
    # def __new__(cls, amount):
    #     return super().__new__(cls, amount)

    if settings.USE_POINTS:
        DECIMAL_PLACES = settings.POINTS_DECIMAL_PLACES

    @classmethod
    def _format_currency(cls, number):
        if settings.USE_POINTS:

            formatted_number = formats.number_format(number)

            if hasattr(settings, 'POINTS_CUSTOM_FORMAT'):
                return settings.POINTS_CUSTOM_FORMAT.format(formatted_number)

            # Translators: display a number of points,
            # like "1 point", "2 points", ...
            # See "Plural-Forms" above for pluralization rules
            # in this language.
            # Explanation at http://bit.ly/1IurMu7
            # In most languages, msgstr[0] is singular,
            # and msgstr[1] is plural
            # the {} represents the number;
            # don't forget to include it in your translation
            return ungettext('{} point', '{} points', number).format(
                formatted_number)
        else:
            return super(Currency, cls)._format_currency(number)

    def to_real_world_currency(self, session):
        if settings.USE_POINTS:
            return RealWorldCurrency(
                float(self) *
                session.config['real_world_currency_per_point'])
        else:
            return self


class _CurrencyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, easymoney.Money):
            if obj.DECIMAL_PLACES == 0:
                return int(obj)
            return float(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def safe_json(obj):
    return mark_safe(json.dumps(obj, cls=_CurrencyEncoder))


def currency_range(first, last, increment):
    assert last >= first
    if Currency(increment) == 0:
        if settings.USE_POINTS:
            setting_name = 'POINTS_DECIMAL_PLACES'
        else:
            setting_name = 'REAL_WORLD_CURRENCY_DECIMAL_PLACES'
        raise ValueError(
            ('currency_range() step argument must not be zero. '
             'Maybe your {} setting is '
             'causing it to be rounded to 0.').format(setting_name)
        )

    assert increment > 0  # not negative

    values = []
    current_value = Currency(first)

    while True:
        if current_value > last:
            return values
        values.append(current_value)
        current_value += increment



from .otree_tags import (
    template, FormFieldNode, NextButtonNode,
    c, abs_value, safe_json,
)

# renaming otree_tags to otree and removing internal tags
# this code is duplicated in otree_tags. I duplicate it rather than importing
# register, because PyCharm's autocomplete doesn't detect the import and
# flags all the template tags in yellow.

register = template.Library()
register.tag('formfield', FormFieldNode.parse)
register.tag('next_button', NextButtonNode.parse)
register.filter(name='c', filter_func=c)
register.filter(name='abs', filter_func=abs_value)
register.filter(name='json', filter_func=safe_json)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from otree.common_internal import (
    get_models_module, in_round, in_rounds)

from otree.db import models
from otree.models.fieldchecks import ensure_field

ATTRIBUTE_ERROR_MESSAGE = '''
Player object has no attribute '{}'. If it is a model field or method,
it must be declared on the Player class in models.py.
'''.replace('\n', ' ')

#from save_the_change.decorators import SaveTheChange
#@SaveTheChange
class BasePlayer(models.Model):
    """
    Base class for all players.
    """

    class Meta:
        abstract = True
        index_together = ['participant', 'round_number']
        ordering = ['pk']

    id_in_group = models.PositiveIntegerField(
        null=True,
        db_index=True,
        doc=("Index starting from 1. In multiplayer games, "
             "indicates whether this is player 1, player 2, etc.")
    )

    # don't modify this directly! Set player.payoff instead
    _payoff = models.CurrencyField(
        null=True,
        doc="""The payoff the player made in this subsession""",
        default=0
    )

    participant = models.ForeignKey(
        'otree.Participant', related_name='%(app_label)s_%(class)s'
    )

    session = models.ForeignKey(
        'otree.Session', related_name='%(app_label)s_%(class)s'
    )

    round_number = models.PositiveIntegerField(db_index=True)

    _gbat_arrived = models.BooleanField(default=False)
    _gbat_grouped = models.BooleanField(default=False)

    def __getattribute__(self, name):
        try:
            return super(BasePlayer, self).__getattribute__(name)
        except AttributeError:
            raise AttributeError(ATTRIBUTE_ERROR_MESSAGE.format(name)) from None

    # it's _name instead of name because people might define
    # their own name field
    def _name(self):
        return self.participant.__unicode__()

    @property
    def payoff(self):
        return self._payoff

    @payoff.setter
    def payoff(self, value):
        if value is None:
            value = 0
        delta = value - self._payoff
        self._payoff += delta
        self.participant.payoff += delta
        # should save it because it may not be obvious that modifying
        # player.payoff also changes a field on a different model
        self.participant.save()

    @property
    def id_in_subsession(self):
        return self.participant.id_in_session

    def __repr__(self):
        id_in_subsession = self.id_in_subsession
        if id_in_subsession < 10:
            # 2 spaces so that it lines up if printing a matrix
            fmt_string = '<Player  {}>'
        else:
            fmt_string = '<Player {}>'
        return fmt_string.format(id_in_subsession)

    def role(self):
        # you can make this depend of self.id_in_group
        return ''

    def in_round(self, round_number):
        return in_round(type(self), round_number, participant=self.participant)

    def in_rounds(self, first, last):
        return in_rounds(type(self), first, last, participant=self.participant)

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number - 1)

    def in_all_rounds(self):
        '''i do it this way because it doesn't rely on idmap'''
        return self.in_previous_rounds() + [self]

    def get_others_in_group(self):
        return [p for p in self.group.get_players() if p != self]

    def get_others_in_subsession(self):
        return [p for p in self.subsession.get_players() if p != self]

    def __unicode__(self):
        return self._name()

    def _GroupClass(self):
        return self._meta.get_field('group').rel.to

    @property
    def _Constants(self):
        return get_models_module(self._meta.app_config.name).Constants

    @classmethod
    def _ensure_required_fields(cls):
        """
        Every ``Player`` model requires a foreign key to the ``Subsession`` and
        ``Group`` model of the same app.
        """
        subsession_model = '{app_label}.Subsession'.format(
            app_label=cls._meta.app_label)
        subsession_field = models.ForeignKey(subsession_model)
        ensure_field(cls, 'subsession', subsession_field)

        group_model = '{app_label}.Group'.format(
            app_label=cls._meta.app_label)
        group_field = models.ForeignKey(group_model, null=True)
        ensure_field(cls, 'group', group_field)




from otree.api import (
    BaseSubsession, BaseGroup, BasePlayer, BaseConstants, models
)

doc = ""

class Constants(BaseConstants):
    name_in_url = 'constants'
    players_per_group = None
    num_rounds = 1
    c_str = 'a'
    c_int = 1
    c_list = [1,2]


class Subsession(BaseSubsession):
    pass


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    pass


# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import views
from .models import Constants
from otree.api import Bot
from ..group_by_arrival_time_round1.tests import GBATMixin


class PlayerBot(Bot, GBATMixin):

    def play_round(self):
        p1, p2 = self.group.get_players()
        assert p1.type == 'A'
        assert p2.type == 'B'

        # are these tests redundant? keep them for now, until i can prove
        # they are not needed
        self.run_assertions()

        yield views.Page1





from __future__ import division, absolute_import
from . import models
from otree.api import Page
from .models import Constants


class HtmlMissingFields(Page):
    def is_displayed(self):
        return self.player.id_in_group == 1

    template_name = 'bots_check_html/NoFields.html'

    form_model = models.Player
    form_fields = ['field_not_in_template']


class HtmlMissingButton(Page):
    def is_displayed(self):
        return self.player.id_in_group == 2

    template_name = 'bots_check_html/NoButton.html'


class TimeoutPage(Page):
    def is_displayed(self):
        return self.player.id_in_group == 3

    template_name = 'bots_check_html/NoFields.html'
    timeout_seconds = 10


page_sequence = [
    HtmlMissingFields,
    HtmlMissingButton,
    TimeoutPage,

]
